From: "Keshavamurthy, Anil S" <anil.s.keshavamurthy@intel.com>
To: "Keshavamurthy, Anil S" <anil.s.keshavamurthy@intel.com>
Cc: Jeff Garzik <jeff@garzik.org>,
linux-kernel@vger.kernel.org, akpm@osdl.org, ak@suse.de,
gregkh@suse.de, muli@il.ibm.com, asit.k.mallick@intel.com,
suresh.b.siddha@intel.com, arjan@linux.intel.com,
ashok.raj@intel.com, shaohua.li@intel.com, davem@davemloft.net
Subject: Re: [Intel-IOMMU 02/10] Library routines for handling pre-allocated pool of objects
Date: Tue, 5 Jun 2007 13:24:33 -0700 [thread overview]
Message-ID: <20070605202433.GA27638@linux-os.sc.intel.com> (raw)
In-Reply-To: <20070604235105.GA13390@linux-os.sc.intel.com>
On Mon, Jun 04, 2007 at 04:51:05PM -0700, Keshavamurthy, Anil S wrote:
> On Mon, Jun 04, 2007 at 07:43:54PM -0400, Jeff Garzik wrote:
> > On Mon, Jun 04, 2007 at 04:06:49PM -0700, Keshavamurthy, Anil S wrote:
> > > On Mon, Jun 04, 2007 at 06:57:14PM -0400, Jeff Garzik wrote:
> > > > you should add logic to free resources here (or queue_work to free the
> > > > resources), if the pool grows beyond a certain size.
> >
> > > Can be added as an add on, testing showed that pool
> > > grows to a certain size and will not grow beyond that
> > > as we tend to reuse the elements.
> >
> > Yes, but is it possible? If no, what part of the code guarantees the
> > pool is limited?
> >
> > We should not merge code that allows the pool to grow without bound.
> > In-house testing certainly never covers all the cases seen in the
> > field, so I wouldn't make too many assumptions based on that. Some
> > vendor will inevitably build a $BigNum system where the IOMMU is very
> > heavily used.
>
> No problem, I can add the code to free the pool element if
> the curr_count ever goes greater than (min_count + 2 * grow_count)
> then bring the curr_count to min_count + grow_count by freeing
> some pool objects.
>
> A patch which will apply to this current patch will follow soon.
Here goes the patch as promised...
------------------------------------
Now adding the logic to shrink the resource pool
and give back the excess pool object's memory space
back to OS.
This patch add's on top of my previous posting.
Signed-off-by: Anil S Keshavamurthy <anil.s.keshavamurthy@intel.com>
---
lib/respool.c | 84 ++++++++++++++++++++++++++++++++++++++++++++--------------
1 file changed, 65 insertions(+), 19 deletions(-)
Index: linux-2.6.22-rc3/lib/respool.c
===================================================================
--- linux-2.6.22-rc3.orig/lib/respool.c 2007-06-05 12:22:26.000000000 -0700
+++ linux-2.6.22-rc3/lib/respool.c 2007-06-05 12:58:01.000000000 -0700
@@ -67,6 +67,7 @@
{
unsigned long flags;
struct list_head *plist = (struct list_head *)vaddr;
+ bool queue_work = 0;
BUG_ON(!vaddr);
BUG_ON(!ppool);
@@ -74,21 +75,74 @@
spin_lock_irqsave(&ppool->pool_lock, flags);
list_add(plist, &ppool->pool_head);
ppool->curr_count++;
+ if (ppool->curr_count > (ppool->min_count +
+ ppool->grow_count * 2))
+ queue_work = 1;
spin_unlock_irqrestore(&ppool->pool_lock, flags);
+
+ if (queue_work)
+ schedule_work(&ppool->work); /* queue work to shrink the pool */
+}
+
+void
+__grow_resource_pool(struct resource_pool *ppool,
+ unsigned int grow_count)
+{
+ unsigned long flags;
+ struct list_head *plist;
+
+ while (grow_count) {
+ plist = (struct list_head *)ppool->alloc_mem(ppool->alloc_size,
+ GFP_KERNEL);
+
+ if (!plist)
+ break;
+
+ /* Add the element to the list */
+ spin_lock_irqsave(&ppool->pool_lock, flags);
+ list_add(plist, &ppool->pool_head);
+ ppool->curr_count++;
+ spin_unlock_irqrestore(&ppool->pool_lock, flags);
+ grow_count--;
+ }
}
+void
+__shrink_resource_pool(struct resource_pool *ppool,
+ unsigned int shrink_count)
+{
+ unsigned long flags;
+ struct list_head *plist;
+
+ while (shrink_count) {
+ /* remove an object from the pool */
+ spin_lock_irqsave(&ppool->pool_lock, flags);
+ if (list_empty(&ppool->pool_head)) {
+ spin_unlock_irqrestore(&ppool->pool_lock, flags);
+ break;
+ }
+ plist = ppool->pool_head.next;
+ list_del(plist);
+ ppool->curr_count--;
+ spin_unlock_irqrestore(&ppool->pool_lock, flags);
+ ppool->free_mem(plist, ppool->alloc_size);
+ shrink_count--;
+ }
+}
+
+
/**
* grow_resource_pool - grows the given resource pool
* @work - work struct
* This functions gets the resource pool pointer from the
* work struct and grows the resource pool by grow_count.
*/
static void
-grow_resource_pool(struct work_struct * work)
+resize_resource_pool(struct work_struct * work)
{
struct resource_pool *ppool;
- struct list_head *plist;
unsigned int min_count, grow_count = 0;
+ unsigned int shrink_count = 0;
unsigned long flags;
ppool = container_of(work, struct resource_pool, work);
@@ -98,22 +152,14 @@
min_count = ppool->min_count + ppool->grow_count;
if (ppool->curr_count < min_count)
grow_count = min_count - ppool->curr_count;
+ else if (ppool->curr_count > min_count + ppool->grow_count)
+ shrink_count = ppool->curr_count - min_count;
spin_unlock_irqrestore(&ppool->pool_lock, flags);
- while(grow_count) {
- plist = (struct list_head *)ppool->alloc_mem(ppool->alloc_size,
- GFP_KERNEL);
-
- if (!plist)
- break;
-
- /* Add the element to the list */
- spin_lock_irqsave(&ppool->pool_lock, flags);
- list_add(plist, &ppool->pool_head);
- ppool->curr_count++;
- spin_unlock_irqrestore(&ppool->pool_lock, flags);
- grow_count--;
- }
+ if (grow_count)
+ __grow_resource_pool(ppool, grow_count);
+ else if (shrink_count)
+ __shrink_resource_pool(ppool, shrink_count);
}
/**
@@ -166,10 +212,10 @@
res->free_mem = free_fn;
spin_lock_init(&res->pool_lock);
INIT_LIST_HEAD(&res->pool_head);
- INIT_WORK(&res->work, grow_resource_pool);
+ INIT_WORK(&res->work, resize_resource_pool);
/* grow the pool */
- grow_resource_pool(&res->work);
+ resize_resource_pool(&res->work);
return (res->curr_count == 0);
}
next prev parent reply other threads:[~2007-06-05 20:28 UTC|newest]
Thread overview: 22+ messages / expand[flat|nested] mbox.gz Atom feed top
2007-06-04 21:02 [Intel-IOMMU 00/10] Intel IOMMU support anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 01/10] DMAR detection and parsing logic anil.s.keshavamurthy
2007-06-04 22:54 ` Jeff Garzik
2007-06-04 22:58 ` Keshavamurthy, Anil S
2007-06-04 23:03 ` Jeff Garzik
2007-06-04 23:17 ` Keshavamurthy, Anil S
2007-06-04 21:02 ` [Intel-IOMMU 02/10] Library routines for handling pre-allocated pool of objects anil.s.keshavamurthy
2007-06-04 22:57 ` Jeff Garzik
2007-06-04 23:06 ` Keshavamurthy, Anil S
2007-06-04 23:43 ` Jeff Garzik
2007-06-04 23:51 ` Keshavamurthy, Anil S
2007-06-05 20:24 ` Keshavamurthy, Anil S [this message]
2007-06-05 20:30 ` Jeff Garzik
2007-06-05 20:48 ` Keshavamurthy, Anil S
2007-06-04 21:02 ` [Intel-IOMMU 03/10] PCI generic helper function anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 04/10] clflush_cache_range now takes size param anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 05/10] IOVA allocation and management routines anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 06/10] Intel IOMMU driver anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 07/10] Intel iommu cmdline option - forcedac anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 08/10] DMAR fault handling support anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 09/10] Iommu Gfx workaround anil.s.keshavamurthy
2007-06-04 21:02 ` [Intel-IOMMU 10/10] Iommu floppy workaround anil.s.keshavamurthy
Reply instructions:
You may reply publicly to this message via plain-text email
using any one of the following methods:
* Save the following mbox file, import it into your mail client,
and reply-to-all from there: mbox
Avoid top-posting and favor interleaved quoting:
https://en.wikipedia.org/wiki/Posting_style#Interleaved_style
* Reply using the --to, --cc, and --in-reply-to
switches of git-send-email(1):
git send-email \
--in-reply-to=20070605202433.GA27638@linux-os.sc.intel.com \
--to=anil.s.keshavamurthy@intel.com \
--cc=ak@suse.de \
--cc=akpm@osdl.org \
--cc=arjan@linux.intel.com \
--cc=ashok.raj@intel.com \
--cc=asit.k.mallick@intel.com \
--cc=davem@davemloft.net \
--cc=gregkh@suse.de \
--cc=jeff@garzik.org \
--cc=linux-kernel@vger.kernel.org \
--cc=muli@il.ibm.com \
--cc=shaohua.li@intel.com \
--cc=suresh.b.siddha@intel.com \
/path/to/YOUR_REPLY
https://kernel.org/pub/software/scm/git/docs/git-send-email.html
* If your mail client supports setting the In-Reply-To header
@@ -74,21 +75,74 @@
/**
- * grow_resource_pool - grows the given resource pool
+ * resize_resource_pool - resize the given resource pool
* @work - work struct
* This functions gets the resource pool pointer from the
via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line
before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox